# coding: utf-8

from enum import Enum
from os import minor
from typing import ClassVar, Tuple, Union
# region Error Classes
class ArgumentError(ValueError):
    '''The error that is raised when one of the arguments provided to a method is not valid.'''
class ArgumentNullError(ValueError):
    '''The error that is raised when a null reference is passed to a method that does not accept it as a valid argument.'''

class ArgumentOutOfRangeError(IndexError):
    '''The error that is raised when the value of an argument is outside the allowable range of values as defined by the invoked method.'''

class FormatError(ValueError):
    '''The error that is raised when the format of an argument is invalid, or when a composite format string is not well formed.'''

    
# endregion Error Classes
class Version(dict):
    '''
    Represents the version number. This class cannot be extended.
    
    Remarks:
        Version instances can be compared usins `=`, `<`, `<=`, `>`, `>=`, `!=`
    '''
    # region Internal Classes
    class ParseFailureKind(Enum):
        NONE = 0
        ARGUMENT_NULL_ERR = 1
        ARGUMENT_ERR = 2
        ARGUMENT_OUT_OF_RANGE_ERR = 3
        FORMAT_ERR = 4

    class VersionResult:
        def __init__(self):
            self._parsed_version = None # Version
            self._failure: Version.ParseFailureKind = Version.ParseFailureKind.NONE # Version.ParseFailureKind
            self._exception_argument = None # str
            self._argument_name = '' # str
            self._can_throw = False # bool

        def init(self, argument_name:str, can_throw:bool):
            self._can_throw = can_throw
            self._argument_name = argument_name
        
        def set_failure(self, failure: 'Version.ParseFailureKind', argument:str = ""):
            self._failure = failure
            self._exception_argument = argument
            if self._can_throw:
                raise self.get_version_parse_exception()

        def get_version_parse_exception(self):
            if self._failure == Version.ParseFailureKind.ARGUMENT_NULL_ERR:
                return ArgumentNullError(self._argument_name)
            if self._failure == Version.ParseFailureKind.ARGUMENT_ERR:
                return ArgumentError("Version string portion was too short or too long.")
            if self._failure == Version.ParseFailureKind.ARGUMENT_OUT_OF_RANGE_ERR:
                return ArgumentOutOfRangeError("Version's parameters must be greater than or equal to zero.")
            if self._failure == Version.ParseFailureKind.FORMAT_ERR:
                try:
                    _ = int(self._exception_argument)
                except ValueError as e:
                    return ArgumentError(e)
                return FormatError("Input string was not in a correct format.")
            return ArgumentError("Version string portion was too short or too long.")
        
        @property
        def parsed_version(self) -> 'Version':
            return self._parsed_version
        
        @parsed_version.setter
        def parsed_version(self, value: 'Version') -> None:
            self._parsed_version = value
    # endregion Internal Classes
    _default_no_val: int = -999
    
    def __init__(self, *args):
        '''
        Class constructor
        @version:
            A string containing the major, minor, build, and revision numbers, where each number is delimited
            with a period character ('.').
        @arg1: Type:int|str If integer then the major version number.
            If str then will be parsed to construct major, minor, build and revision. Samse as `Version.parse()`
        @arg2: Type:int|str, the minor version number.
        @arg3: Type:int|str, the build number.
        @arg4: Type:int|str, the revision number.
        
        Remarks:
            The version parameter can contain only the components major, minor, build, and revision, in that
            order, and all separated by periods. There must be at least one component, and at most four. The
            first two components are assumed to be major and minor. The value of unspecified components is
            `0`.
            The format of the version number is as follows. Optional components are shown in square brackets
            ('[' and ']'):
            major.minor[.build[.revision]]
            All defined components must be integers greater than or equal to 0. For example, if the major number
            is 3, the minor number is 4, the build number is 2, and the revision number is 5, then version
            should be "3.4.2.5".
        
        Version(major, minor)
            Initializes a new instance of the MfVersion class using the specified major and minor values.
      
        Version(major, minor, build)
            Initializes a new instance of the MfVersion class using the specified major, minor, and build values.
        
        Version(major, minor, build, revision)
            Initializes a new instance of the MfVersion class with the specified major, minor, build, and revision numbers.
        '''
        if self.__class__.__name__ != 'Version' :
            raise TypeError("version is a seal class")
        super_args = {}
        self._major = None
        self._minor = None
        self._build = Version._default_no_val
        self._revision = Version._default_no_val
        self._arg_len = len(args)
        if self._arg_len > 4:
            raise ArgumentError(f"{self.__class__.__name__} has a max of four args")
        
        major:Union[str,int, None] = None
        minor: Union[int, str, float, None] = None
        build: Union[int, str, float, None] = None
        revision: Union[int, str, float, None] = None
        if self._arg_len == 0:
            self._major = 0
            self._minor = 0
            return None
        if self._arg_len >= 1:
            major = args[0]
        if self._arg_len >= 2:
            minor = args[1]
        if self._arg_len >= 3:
            build = args[2]
        if self._arg_len == 4:
            revision = args[3]

        if minor is None:
            if isinstance(major, int):
                self._major = major
                self._minor = 0
                self._set_super_args(super_args, 'major', self._major)
                self._set_super_args(super_args, 'minor', self._minor)
            elif isinstance(major, dict):
                self._init_by_dict(major, super_args)
            else:
                version2 = Version.parse(major)
                self._major= version2._major
                self._minor= version2._minor
                self._build= version2._build
                self._revision = version2._revision
                self._set_super_args(super_args, 'major', self._major)
                self._set_super_args(super_args, 'minor', self._minor)
                if self._build != Version._default_no_val:
                    self._set_super_args(
                        super_args, 'build', self._build)
                if self._revision != Version._default_no_val:
                    self._set_super_args(
                        super_args, 'revision', self._revision)
        elif build is None:
            self._major = Version._parse_int_arg(arg=major, arg_name='major')
            self._minor = Version._parse_int_arg(arg=minor, arg_name='minor')
            self._set_super_args(super_args, 'major', self._major)
            self._set_super_args(super_args, 'minor', self._minor)
        elif revision is None:
            self._major = Version._parse_int_arg(arg=major, arg_name='major')
            self._minor = Version._parse_int_arg(arg=minor, arg_name='minor')
            self._build = Version._parse_int_arg(arg=build, arg_name='build')
            self._set_super_args(super_args, 'major', self.major)
            self._set_super_args(super_args, 'minor', self._minor)
            self._set_super_args(super_args, 'build', self._build)
        else:
            self._major = Version._parse_int_arg(arg=major, arg_name='major')
            self._minor = Version._parse_int_arg(arg=minor, arg_name='minor')
            self._build = Version._parse_int_arg(arg=build, arg_name='build')
            self._revision = Version._parse_int_arg(arg=revision, arg_name='revision')
            self._set_super_args(super_args, 'major', self.major)
            self._set_super_args(super_args, 'minor', self._minor)
            self._set_super_args(super_args, 'build', self._build)
            self._set_super_args(super_args, 'revision', self._revision)
        self._validate()
        dict.__init__(self, **super_args)
 
    def _set_super_args(self, args:dict, key:str, value:int) -> None:
        if not key in args:
            args[key] = value
            
    def _init_by_dict(self, d:dict, super_args:dict):
        local_d = {
            'major': True,
            'minor':True,
            'build': False,
            'revision': False
        }
        for k, v in local_d.items():
            if k in d:
                val = d[k]
                i = Version._parse_int_arg(val,k)
                setattr(self, f"_{k}",i)
                super_args[k] = i
            else:
                if v:
                    raise ArgumentNullError(f"{self.__class__.__name__} dictionary missing required key of '{k}'")

    @classmethod
    def _parse_int_arg(cls, arg, arg_name:str) -> int:
        if isinstance(arg, int):
            return arg
        if isinstance(arg, str):
            _arg = arg.strip()
            result = -1
            try:
                result = int(_arg)
                return result
            except:
                pass
            _arg = _arg.lower()
            if _arg.startswith('0x'):
                try:
                    result = int(_arg, 16)
                    return result
                except:
                    pass
        msg = (f"{cls.__name__}, arg '{arg_name}' cannot be converted to an interger."
               f" '{arg_name}' type '{type(arg).__name__}'")
        raise ArgumentError(msg)

    def _validate(self) -> None:
        msg = "Version's parameters must be greater than or equal to zero."
        if self._major < 0:
            raise ArgumentOutOfRangeError("major", msg)
        if self._minor < 0:
            raise ArgumentOutOfRangeError("minor", msg)
        if self._build != Version._default_no_val and self._build < 0:
            raise ArgumentOutOfRangeError("build", msg)
        if self._revision != Version._default_no_val and self._revision < 0:
            raise ArgumentOutOfRangeError("revision", msg)
            
    @staticmethod
    def _try_parse_component(component:str, component_name: str, result:'Version.VersionResult') -> Tuple[bool, int]: 
        pc = -1
        try:
            pc = Version._parse_int_arg(component, component_name)
        except Exception as e:
            result.set_failure(Version.ParseFailureKind.FORMAT_ERR, component)
            return False, -1
        if pc < 0:
            result.set_failure(Version.ParseFailureKind.ARGUMENT_OUT_OF_RANGE_ERR, component_name)
            return False , -1
        return True, pc
    
    @staticmethod
    def _try_parse_version(version:str, result: 'Version.VersionResult') -> bool:
        _ver = version.strip()
        if len(_ver) == 0:
            result.set_failure(Version.ParseFailureKind.ARGUMENT_NULL_ERR)
            return False
        ver_lst = _ver.split('.')
        split_count = len(ver_lst)
        if split_count < 2 or split_count > 4:
            result.set_failure(Version.ParseFailureKind.ARGUMENT_ERR)
            return False
        major = 0
        tp = Version._try_parse_component(ver_lst[0], "Version", result)
        if tp[0] == False:
            return False
        else:
            major = tp[1]
        minor = 0
        tp = Version._try_parse_component(ver_lst[1], "Version", result)
        if tp[0] == False:
            return False
        else:
            minor = tp[1]
        split_count -= 2
        if split_count > 0:
            build = 0
            tp = Version._try_parse_component(ver_lst[2],"build", result)
            if tp[0] == False:
                return False
            else:
                build = tp[1]
            split_count -= 1
            if split_count > 0:
                revision = 0
                tp = Version._try_parse_component(ver_lst[3],"revision", result)
                if tp[0] == False:
                    return False
                else:
                    revision = tp[1]
                result.parsed_version = Version(major, minor, build, revision)
            else:
                result.parsed_version = Version(major, minor, build)
        else:
            result.parsed_version = Version(major, minor)
        return True
    
    @staticmethod
    def parse(input: str):
        '''
        Converts the string representation of a version number to an equivalent `Version` instance.
        @input: A string that contains a version number to convert.
        @return: A `Version` instance that is equivalent to the version number specified in the input parameter
        @error: Will raise errors if they occur. 
        @see also: `try_parse()`
        @example:
        ```
        ver = Version.parse("1.3")
        print(ver.major) # 1
        print(ver.minor) # 3
        
        ver = Version.parse("1.3.8.97")
        print(ver.major) # 1
        print(ver.minor) # 3
        print(ver.build) # 8
        print(ver.revision) # 97
        ```
        '''
        if not isinstance(input, str):
            raise ArgumentError("Version.Parse() 'input' must be of type 'str'")
        version_result = Version.VersionResult()
        version_result.init('input', True)
        # no need to test the result of the following line. It will never raise an error when init is set with True: init('input' True)
        Version._try_parse_version(input, version_result)
        return version_result.parsed_version

    @staticmethod
    def try_parse(input: str) -> Tuple[bool, Union['Version',Exception]]:
        '''
        Converts the string representation of a version number to an equivalent `Version` instance.
        @input: A string that contains a version number to convert.
        @return: tuple with the first element as bool. first element will be `True` when parse is a success;
        Otherwise, first element will be `False`.
        When first element is `True` an instance of the `Version` will be returned as the second element.
        When first element is `False` second element will contain the error that occured that caused the failure.
        @example:
        ```
        v_result = Version.try_parse('2.1.12')
        if v_result[0] == True:
            v = v_result[1]
            print(v.major) # 2
            print(v.minor) # 1
            print(v.build) # 17
        else:
            print("An Error has occured", v_result[1])
        ```
        '''
        err = None
        if not isinstance(input, str):
            err = ArgumentError(
                "Version.Parse() 'input' must be of type 'str'")
            return False, err
        version_result = Version.VersionResult()
        version_result.init('input', False)
        if Version._try_parse_version(input, version_result) == False:
            err = version_result.get_version_parse_exception()
            return False, err
        return True, version_result.parsed_version
    # region Properties
    @property
    def build(self) -> int:
        '''
        Gets the value of the build component of the version number for the current Version instance.
        @return: int representing build
        '''
        if self._build == Version._default_no_val:
            return 0
        return self._build
    
    @property
    def major(self)-> int:
        '''
        Gets the value of the major component of the version number for the current Version instnace.
        @return: int representing major
        '''
        return self._major
    
    @property
    def major_revision(self) -> int:
        '''
        Gets the high 16 bits of the revision number.
        
        @remarks:
        Read-only property
        Suppose you release an interim version of your application to temporarily correct a problem until you
        can release a permanent solution. The temporary version does not warrant a new revision number,
        but it does need to be identified as a different version. In this case, encode the identification
        information in the high and low 16-bit portions of the 32-bit revision number. Use the `revision`
        property to obtain the entire revision number, use the `major_revision` property to obtain the high 16 bits,
        and use the `minor_revision` property to obtain the low 16 bits.
        '''
        hi_word = self.revision >> 16
        return hi_word
    
    @property
    def minor(self)-> int:
        '''
        Gets the value of the minor component of the version number for the current Version instnace.
        @return: int representing minor
        '''
        return self._minor
    
    @property
    def minor_revision(self) -> int:
        '''
        Gets the low 16 bits of the revision number.
        
        @remarks:
        Read-only property
        Suppose you release an interim version of your application to temporarily correct a problem until you
        can release a permanent solution. The temporary version does not warrant a new revision number,
        but it does need to be identified as a different version. In this case, encode the identification
        information in the high and low 16-bit portions of the 32-bit revision number. Use the `revision`
        property to obtain the entire revision number, use the `major_revision` property to obtain the high 16 bits,
        and use the `minor_revision` property to obtain the low 16 bits.
        '''
        lo_word = self.revision & int('0xFFFF', 16)
        return lo_word

    @property
    def revision(self)-> int:
        '''
        Gets the value of the revision component of the version number for the current Version instnace.
        @return: int representing minor
        '''
        if self._revision == Version._default_no_val:
            return 0
        return self._revision
    # endregion Properties

    # region Compare

    def _compare(self, other: 'Version'):
        if self.major != other.major:
            if self.major > other.major:
                return 1
            return -1
        elif self.minor != other.minor:
            if self.minor > other.minor:
                return 1
            return -1
        elif self.build != other.build:
            if self.build > other.build:
                return 1
            return -1
        elif self.revision == other.revision:
            return 0
        elif self.revision > other.revision:
            return 1
        return -1


    def __eq__(self, other: 'Version'):
        if not isinstance(other, Version):
            return NotImplemented
        return self._compare(other) == 0

    def __le__(self, other: 'Version'):
        if not isinstance(other, Version):
            return NotImplemented
        return self._compare(other) <= 0

    def __gt__(self, other: 'Version'):
        if not isinstance(other, Version):
            return NotImplemented
        return self._compare(other) > 0

    def __lt__(self, other: 'Version'):
        if not isinstance(other, Version):
            return NotImplemented
        return self._compare(other) < 0

    def __ge__(self, other: 'Version'):
        if not isinstance(other, Version):
            return NotImplemented
        return self._compare(other) >= 0
    
    def __hash__(self):
        _int = 0
        _int |= (self.major & 15) << 28
        _int |= (self.minor & 255) << 20
        _int |= (self.build & 255) << 12
        _int |= (self.revision & 4095)
        return _int
    # endregion Compare
    # region string methodos
    def _to_str(self, field_count:int):
        if field_count == 1:
            return f"{self.major}"
        elif field_count == 2:
            return f"{self.major}.{self.minor}"
        if self._build == Version._default_no_val:
            msg = f"Parameter must be between '0' and '2'."
            raise ArgumentError(msg, 'field_count')
        if field_count == 3:
            return f"{self.major}.{self.minor}.{self.build}"
        if self._revision == Version._default_no_val:
            msg = f"Parameter must be between '0' and '3'."
            raise ArgumentError(msg, 'field_count')
        if field_count == 4:
            return f"{self.major}.{self.minor}.{self.build}.{self.revision}"
        
    def to_str(self, field_count:int=-1) -> str:
        if field_count > 0:
            if field_count > 4:
                raise ArgumentOutOfRangeError("field_count must be from 1 to 4")
            return self._to_str(field_count)
        if self._build == Version._default_no_val:
            return self._to_str(2)
        if self._revision == Version._default_no_val:
            return self._to_str(3)
        return self._to_str(4)
    
    
    def __str__(self):
       return self.to_str()
   
    def __repr__(self):
        s = self.to_str()
        s_arg = s.replace('.', ', ')
        return f"<{self.__class__.__name__}({s_arg})>"
    # endregion string methods



# coding: utf-8
try:
    import path_imports
except:
    pass
import unittest
from src.verr import Version, ArgumentError, ArgumentOutOfRangeError, ArgumentNullError

# run test in order
unittest.TestLoader.sortTestMethodsUsing = None


class TestVersion(unittest.TestCase):

    def test_001(self):
        v = Version()
        self.assertEqual(v.major, 0)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.build, 0)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '0.0')
        self.assertEqual(v.__hash__(), 0)

    def test010_instance_major_int(self):
        v = Version(2)
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.build, 0)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '2.0')
        self.assertEqual(v.__hash__(), 536870912)

    def test011_instance_major_str(self):
        v = Version("2.0")
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.build, 0)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '2.0')
        self.assertEqual(v.__hash__(), 536870912)

    def test012_instance_major_str_short(self):
        self.assertRaises(ArgumentError, Version, "2")

    def test013_instance_major_str_empty(self):
        self.assertRaises(ArgumentNullError, Version, "")

    def test014_instance_major_str_empty_part(self):
        self.assertRaises(ArgumentError, Version, "2..0")

    def test015_instance_major_str_empty_part(self):
        self.assertRaises(ArgumentError, Version, 1, 2, 3, 4, 5)

    def test016_minor_neg(self):
        self.assertRaises(ArgumentOutOfRangeError, Version, 1, 2, -2)

    def test017_rev_neg(self):
        self.assertRaises(ArgumentOutOfRangeError, Version, 1, 2, 2, -1)
        
    def test_018(self):
        self.assertRaises(ArgumentError, Version, "1.a.2")

    def test_019(self):
        self.assertRaises(ArgumentError, Version, 1, 'a',2)
        self.assertRaises(ArgumentError, Version, "1", 'a', 2)
        
    def test020_instance_all_int(self):
        v = Version(1, 2, 3, 4)
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 3)
        self.assertEqual(v.revision, 4)
        self.assertEqual(str(v), '1.2.3.4')
        self.assertEqual(v.__hash__(), 270544900)

    def test021_instance_all_single_str(self):
        v = Version("1.2.3.4")
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 3)
        self.assertEqual(v.revision, 4)
        self.assertEqual(str(v), '1.2.3.4')

    def test022_instance_m_m_b(self):
        v = Version("1.2.3")
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 3)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '1.2.3')

    def test023_instance_m_m(self):
        v = Version("1.2")
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 0)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '1.2')

    def test024_instance_int_m_m_b(self):
        v = Version(1,2,44)
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 44)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '1.2.44')
        self.assertEqual(v.__hash__(), 270712832)

    def test025_instance_int_m_m(self):
        v = Version(24, 29)
        self.assertEqual(v.major, 24)
        self.assertEqual(v.minor, 29)
        self.assertEqual(v.build, 0)
        self.assertEqual(v.revision, 0)
        
    def test030_parse(self):
        v = Version.parse('0.1.0')
        self.assertEqual(v.major, 0)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.build, 0)

    def test031_parse_all(self):
        v = Version.parse("1.2.3.4")
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 3)
        self.assertEqual(v.revision, 4)

    def test031_parse_too_many(self):
        self.assertRaises(ArgumentError, Version.parse, "1.2.3.4.5")

    def test031_parse_too_few(self):
        self.assertRaises(ArgumentError, Version.parse, "1")

    def test031_parse_missing_part(self):
        self.assertRaises(ArgumentError, Version.parse, "2.1.0.")

    def test032_parse_empty_str(self):
        self.assertRaises(ArgumentNullError, Version.parse, "")

    def test033parse_empty_none(self):
        self.assertRaises(ArgumentError, Version.parse, None)

    def test034parse_neg(self):
        self.assertRaises(ArgumentOutOfRangeError, Version.parse, "1.-2")
    
    def test035parse_empty_str(self):
        self.assertRaises(ArgumentNullError, Version.parse, "")

    def test050_eq(self):
        v1 = Version.parse("1.2.3.4")
        v2 = Version.parse("1.2.3.4")
        self.assertTrue(v1 == v2)
        self.assertFalse(v1 == 22)

    def test052_rev_gt_eq(self):
        v1 = Version.parse("1.2.3")
        v2 = Version.parse("1.2.3")
        self.assertTrue(v1 >= v2)
        _type_error = False
        try:
            print(v1 >= 22)
        except TypeError:
            _type_error = True
        self.assertTrue(_type_error)

    def test053_rev_gt_eq(self):
        v1 = Version.parse("1.2.3")
        v2 = Version.parse("1.2.3")
        self.assertTrue(v1 <= v2)
        _type_error = False
        try:
            print(v1 <= 22)
        except TypeError:
            _type_error = True
        self.assertTrue(_type_error)

    def test054_rev_gt(self):
        v1 = Version.parse("1.2.3.2")
        v2 = Version.parse("1.2.3")
        self.assertTrue(v1 > v2)
        _type_error = False
        try:
            print(v1 > 22)
        except TypeError:
            _type_error = True
        self.assertTrue(_type_error)

    def test054_rev_lt(self):
        v1 = Version("1.2.2")
        v2 = Version("1.2.3")
        self.assertTrue(v1 < v2)
        _type_error = False
        try:
            print(v1 < 22)
        except TypeError:
            _type_error = True
        self.assertTrue(_type_error)

    def test055_rev_ne(self):
        v1 = Version("1.2.2")
        v2 = Version("1.2.3")
        self.assertTrue(v1 != v2)
        self.assertTrue(v1 != 22)

    def test056_major_comp(self):
        v1 = Version("2.2.2")
        v2 = Version("1.2.2")
        self.assertTrue(v1 > v2)
        self.assertTrue(v2 < v1)
        
    def test057_minor_comp(self):
        v1 = Version("1.3.2")
        v2 = Version("1.2.2")
        self.assertTrue(v1 > v2)
        self.assertTrue(v2 < v1)
        
    def test058_build_comp(self):
        v1 = Version("1.2.3")
        v2 = Version("1.2.2")
        self.assertTrue(v1 > v2)
        self.assertTrue(v2 < v1)

    def test059_rev_comp(self):
        v1 = Version("1.2.2.2")
        v2 = Version("1.2.2.1")
        self.assertTrue(v1 > v2)
        self.assertTrue(v2 < v1)


    def test060_str_eq(self):
        v1 = Version("1.2.2")
        self.assertEqual(str(v1), "1.2.2")

    def test070_str_neg_major(self):
        self.assertRaises(ArgumentOutOfRangeError, Version, -1, 2)
        
    def test071_str_neg_minor(self):
        self.assertRaises(ArgumentOutOfRangeError, Version, 1, -2)
        
    def test072_str_neg_build(self):
        self.assertRaises(ArgumentOutOfRangeError, Version, "1.2.-3")
    
    def test072_str_neg_rev(self):
        self.assertRaises(ArgumentOutOfRangeError, Version, "1.2.3.-5")
    
    def test080_to_str(self):
        v = Version("1.2.3.4")
        self.assertRaises(ArgumentOutOfRangeError, v.to_str, 5)
        self.assertEqual(v.to_str(1), '1')
        self.assertEqual(v.to_str(2), '1.2')
        self.assertEqual(v.to_str(3), '1.2.3')
        self.assertEqual(v.to_str(4), '1.2.3.4')
    
    def test081_to_str_build_err(self):
        v = Version("1.2")
        self.assertRaises(ArgumentOutOfRangeError, v.to_str, 5)
        self.assertEqual(v.to_str(1), '1')
        self.assertEqual(v.to_str(2), '1.2')
        self.assertRaises(ArgumentError, v.to_str, 3)
    
    def test082_to_str_rev_err(self):
        v = Version("1.2.3")
        self.assertRaises(ArgumentOutOfRangeError, v.to_str, 5)
        self.assertEqual(v.to_str(1), '1')
        self.assertEqual(v.to_str(2), '1.2')
        self.assertEqual(v.to_str(3), '1.2.3')
        self.assertRaises(ArgumentError, v.to_str, 4)

    def test090_rev_hi_lo(self):
        rev = int("ffbbff0b", 16)
        v = Version(1,2,3,rev)
        self.assertEqual(v.major_revision, int("ffbb", 16))
        self.assertEqual(v.minor_revision, int("ff0b", 16))
    def test100_inherit(self):
        class tmp(Version):
            pass
        self.assertRaises(TypeError, tmp, "1.2.-3")

    def test200_try_parse(self):
        v_result = Version.try_parse('0.1.0')
        v = v_result[1]
        self.assertTrue(v_result[0])
        self.assertEqual(v.major, 0)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.build, 0)
        
    def test201_try_parse_empty_str(self):
        v_result = Version.try_parse('')
        e = v_result[1]
        self.assertFalse(v_result[0])
        self.assertIsInstance(e, ArgumentNullError)
    
    def test202_try_parse_non_str(self):
        v_result = Version.try_parse(2)
        e = v_result[1]
        self.assertFalse(v_result[0])
        self.assertIsInstance(e, ArgumentError)
        
    def test203_try_parse(self):
        v_result = Version.try_parse('0.1.0.2.1')
        e = v_result[1]
        self.assertIsInstance(e, ArgumentError)

    def test204_try_parse(self):
        v_result = Version.try_parse('1')
        e = v_result[1]
        self.assertIsInstance(e, ArgumentError)

    def test205_try_parse_neg_minor(self):
        v_result = Version.try_parse('1.-2')
        e = v_result[1]
        self.assertIsInstance(e, ArgumentOutOfRangeError)

    def test206_try_parse_neg_build(self):
        v_result = Version.try_parse('1.2.-3')
        e = v_result[1]
        self.assertIsInstance(e, ArgumentOutOfRangeError)

    def test207_try_parse_neg_rev(self):
        v_result = Version.try_parse('1.2.3.-5')
        e = v_result[1]
        self.assertIsInstance(e, ArgumentOutOfRangeError)
    
    def test208_try_parse_bad_build(self):
        v_result = Version.try_parse('a.2.3.5')
        e = v_result[1]
        self.assertIsInstance(e, ArgumentError)

    def test209_try_parse_hex(self):
        s = "0x1f.0x4D.0XAa.0xbb"
        v_result = Version.try_parse(s)
        v = v_result[1]
        self.assertTrue(v_result[0])
        self.assertEqual(v.major, 31)
        self.assertEqual(v.minor, 77)
        self.assertEqual(v.build, 170)
        self.assertEqual(v.revision, 187)
    
    def test210_try_parse_bad_hex(self):
        s = "0x1f.0e4D.0XAa.0xbb"
        v_result = Version.try_parse(s)
        e = v_result[1]
        self.assertFalse(v_result[0])
        self.assertIsInstance(e, ArgumentError)

    def test301_parse_hex(self):
        s = "0x1f.0x4D.0XAa.0xbb"
        v = Version.parse(s)
        self.assertEqual(v.major, 31)
        self.assertEqual(v.minor, 77)
        self.assertEqual(v.build, 170)
        self.assertEqual(v.revision, 187)
    
    def test303_parse_hex_with_bad_value(self):
        s = "0x1f.0x4D.0XAa.0ybb"
        self.assertRaises(ArgumentError, Version.parse,s)
    
    def test303_parse_non_hex(self):
        s = "0x1f.0x4D.0XAa.0xbg"
        self.assertRaises(ArgumentError, Version.parse, s)

if __name__ == '__main__':
    unittest.main()



try:
    import path_imports
except:
    pass
import unittest
from src.verr import Version, ArgumentError, ArgumentOutOfRangeError, ArgumentNullError
import json

# run test in order
unittest.TestLoader.sortTestMethodsUsing = None


class TestVersionJson(unittest.TestCase):
    def test_001(self):
        v = Version()
        self.assertEqual(v.major, 0)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.build, 0)
        self.assertEqual(v.revision, 0)
        self.assertEqual(str(v), '0.0')
        self.assertEqual(v.__hash__(), 0)

    def test_010_json_m_m_b_r(self):
        v = Version("1.2.3.4")
        j = json.dumps(v)
        self.assertEqual(
            j, '{"major": 1, "minor": 2, "build": 3, "revision": 4}')
        v_j = Version(json.loads(j))
        self.assertEqual(v_j.major, 1)
        self.assertEqual(v_j.minor, 2)
        self.assertEqual(v_j.build, 3)
        self.assertEqual(v_j.revision, 4)
        self.assertEqual(v_j.__str__(), '1.2.3.4')
        self.assertEqual(v_j.__repr__(), '<Version(1, 2, 3, 4)>')

    def test_011_json_m_m_b(self):
        v = Version("1.2.3")
        j = json.dumps(v)
        self.assertEqual(j, '{"major": 1, "minor": 2, "build": 3}')
        v_j = Version(json.loads(j))
        self.assertEqual(v_j.major, 1)
        self.assertEqual(v_j.minor, 2)
        self.assertEqual(v_j.build, 3)
        self.assertEqual(v_j.revision, 0)

    def test_012_json_m_m(self):
        v = Version.parse("1.2")
        j = json.dumps(v)
        self.assertEqual(j, '{"major": 1, "minor": 2}')
        v_j = Version(json.loads(j))
        self.assertEqual(v_j.major, 1)
        self.assertEqual(v_j.minor, 2)
        self.assertEqual(v_j.build, 0)
        self.assertEqual(v_j.revision, 0)

    def test_013_json_m_errro(self):
        j = '{"major": 1}'
        d = json.loads(j)
        self.assertRaises(ArgumentNullError, Version, d)

    def test_014_json_m_m_b_r_more(self):
        j = '{"major": 1, "minor": 2, "build": 3, "revision": 4, "year": 2021}'
        v_j = Version(json.loads(j))
        self.assertEqual(v_j.major, 1)
        self.assertEqual(v_j.minor, 2)
        self.assertEqual(v_j.build, 3)
        self.assertEqual(v_j.revision, 4)
        self.assertEqual(v_j.__str__(), '1.2.3.4')
        self.assertEqual(v_j.__repr__(), '<Version(1, 2, 3, 4)>')

    def test_015_json_m_m_b_repeat(self):
        v = Version("1.2.3")
        j = json.dumps(v)
        self.assertEqual(j, '{"major": 1, "minor": 2, "build": 3}')
        v_j = Version(json.loads(j))
        self.assertEqual(v_j.major, 1)
        self.assertEqual(v_j.minor, 2)
        self.assertEqual(v_j.build, 3)
        self.assertEqual(v_j.revision, 0)
        self.assertTrue(v == v_j)
        self.assertEqual(v.__hash__(), v_j.__hash__())

        j = json.dumps(v_j)
        v = Version(json.loads(j))
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 3)
        self.assertEqual(v.revision, 0)
        self.assertTrue(v == v_j)
        self.assertEqual(v.__hash__(), v_j.__hash__())

    def test_020_json_missing_major(self):
        j = '{"minor": 2, "build": 3, "revision": 4}'
        d = json.loads(j)
        self.assertRaises(ArgumentNullError, Version, d)

    def test_021_json_missing_minor(self):
        j = '{"major": 2, "build": 3, "revision": 4}'
        d = json.loads(j)
        self.assertRaises(ArgumentNullError, Version, d)

    def test_022_json_neg_major(self):
        j = '{"major": -1, "minor": 2, "build": 3, "revision": 4}'
        d = json.loads(j)
        self.assertRaises(ArgumentOutOfRangeError, Version, d)

    def test_023_json_neg_minor(self):
        j = '{"major": 1, "minor": -2, "build": 3, "revision": 4}'
        d = json.loads(j)
        self.assertRaises(ArgumentOutOfRangeError, Version, d)

    def test_024_json_neg_build(self):
        j = '{"major": 1, "minor": 2, "build": -3, "revision": 4}'
        d = json.loads(j)
        self.assertRaises(ArgumentOutOfRangeError, Version, d)

    def test_025_json_neg_revision(self):
        j = '{"major": 1, "minor": 2, "build": 3, "revision": -4}'
        d = json.loads(j)
        self.assertRaises(ArgumentOutOfRangeError, Version, d)

    def test_030_json_str_values(self):
        j = '{"major": "1", "minor": "2", "build": "3", "revision": "4"}'
        v = Version(json.loads(j))
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.build, 3)
        self.assertEqual(v.revision, 4)
        k = '{"major": 1, "minor": 2, "build": 3, "revision": 4}'
        v_dump = json.dumps(v)
        self.assertEqual(k, v_dump)

    def test_031_json_str_bad_val(self):
        j = '{"major": "1", "minor": "b", "build": "3", "revision": "4"}'
        d = json.loads(j)
        self.assertRaises(ArgumentError, Version, d)

    def test_032_json_str_empty_val(self):
        j = '{"major": "1", "minor": "3", "build": "", "revision": "4"}'
        d = json.loads(j)
        self.assertRaises(ArgumentError, Version, d)

if __name__ == '__main__':
    unittest.main()



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
with open(HERE/ "README.md") as fh:
    README = fh.read()

# This call to setup() does all the work
setup(
    name="verr",
    version="0.1.3",
    description="Represents a version number that can be parsed and compared.",
    long_description_content_type="text/markdown",
    long_description=README,
    url="https://github.com/Amourspirit/python-version-num",
    author=":Barry-Thomas-Paul: Moss",
    license="MIT",
    package_dir={'': 'src'},
    py_modules=['verr'],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
)
